/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid.smartdocuments;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.Optional;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.http.HttpRequest;
import org.springframework.http.HttpStatusCode;
import org.springframework.http.MediaType;
import org.springframework.http.client.ClientHttpResponse;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestClient;
import org.w3c.dom.Document;
import org.w3c.dom.Text;
import org.xml.sax.SAXException;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;

import de.ozgcloud.document.bescheid.BescheidResponse;
import de.ozgcloud.document.bescheid.BescheidRemoteService;
import de.ozgcloud.document.bescheid.BescheidRequest;
import de.ozgcloud.document.bescheid.common.user.UserProfile;
import de.ozgcloud.document.bescheid.smartdocuments.SmartDocumentsBescheidRemoteService.SmartDocumentsResponse.SmartDocumentDocument;
import de.ozgcloud.document.bescheid.smartdocuments.SmartDocumentsBescheidRemoteService.SmartDocumentsResponse.SmartDocumentFile;
import de.ozgcloud.document.bescheid.smartdocuments.SmartDocumentsRequest.CustomerData;
import de.ozgcloud.document.bescheid.smartdocuments.SmartDocumentsRequest.CustomerData.BescheidData;
import de.ozgcloud.document.bescheid.smartdocuments.SmartDocumentsRequest.CustomerData.UserData;
import de.ozgcloud.document.bescheid.smartdocuments.SmartDocumentsRequest.SmartDocument.Selection;
import de.ozgcloud.document.bescheid.vorgang.Vorgang;
import de.ozgcloud.common.binaryfile.FileDataDeserializer;
import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
@RequiredArgsConstructor
@ConditionalOnProperty("ozgcloud.bescheid.smart-documents.url")
class SmartDocumentsBescheidRemoteService implements BescheidRemoteService {

	private static final String FILE_TYPE_PDF = "PDF";
	private static final String FILE_TYPE_XML = "XML";

	static final String NACHRICHTEN_TEXT_EXPRESSION = "/root/SmartDocument/Fields/NachrichtenText/text()"; // NOSONAR
	static final String FIELD_TEMPLATE_EXPRESSION = "/root/SmartDocument/Fields/Field[@ID=\"Template.56E7AA0956C7486292E9A02114CB231C\"]/text()"; // NOSONAR

	@SuppressWarnings("deprecation") // SD requires forced UTF-8 encoding
	private static final MediaType JSON_MEDIA_TYPE_FOR_SD = MediaType.APPLICATION_JSON_UTF8;

	@Qualifier("smartDocuments")
	private final RestClient smartDocumentsRestClient;

	private final SmartDocumentsProperties properties;

	@Override
	public BescheidResponse create(BescheidRequest request, Vorgang vorgang) {
		var sdRequest = createRequest(request, vorgang);
		LOG.debug(() -> buildLogRequest(sdRequest));

		var response = smartDocumentsRestClient.post().accept(MediaType.APPLICATION_JSON)
				.contentType(JSON_MEDIA_TYPE_FOR_SD)
				.body(sdRequest)
				.retrieve()
				.onStatus(HttpStatusCode::is4xxClientError, this::handleClientError)
				.toEntity(SmartDocumentsResponse.class)
				.getBody();
		return buildBescheid(request, response);

	}

	void handleClientError(HttpRequest request, ClientHttpResponse response) {
		try {
			var responseBody = IOUtils.toString(response.getBody(), StandardCharsets.UTF_8);
			throw new TechnicalException("Client-Error: " + response.getStatusCode() + " " + responseBody);
		} catch (IOException e) {
			throw new TechnicalException("Couldn't read response body", e);
		}
	}

	private String buildLogRequest(SmartDocumentsRequest request) {
		try {
			ObjectMapper oj = new ObjectMapper();
			return oj.writerWithDefaultPrettyPrinter().writeValueAsString(request);
		} catch (Exception e) {
			LOG.warn("Error writing request as json-string for logging.", e);
			return "Error generating logging string";
		}
	}

	BescheidResponse buildBescheid(BescheidRequest request, SmartDocumentsResponse smartDocumentsResponse) {
		var smartDocumentsFile = getSmartDocumentsFile(smartDocumentsResponse);

		return BescheidResponse.builder()
				.bescheidFile(smartDocumentsFile.getDocument().getData())
				.bescheidFileName(smartDocumentsFile.getFilename())
				.size(smartDocumentsFile.getDocument().getData().length())
				.contentType(MediaType.APPLICATION_PDF_VALUE)
				.nachrichtText(getNachrichtText(smartDocumentsResponse))
				.bewilligt(request.isGenehmigt())
				.createdBy(request.getCreateFor().getId())
				.vorgangId(request.getVorgangId())
				.build();
	}

	private SmartDocumentFile getSmartDocumentsFile(SmartDocumentsResponse response) {
		return getSmartDocumentsFile(response, FILE_TYPE_PDF)
				.orElseThrow(() -> new IllegalStateException("No PDF File in SmartDocuments Response found."));
	}

	Optional<String> getNachrichtText(SmartDocumentsResponse response) {
		return getXMLFile(response).flatMap(this::extractTextFromXmlFile);
	}

	Optional<String> extractTextFromXmlFile(File xmlFile) {
		try {
			var document = parseXml(xmlFile);
			if (document.isPresent()) {
				return document.flatMap(this::getNachrichtenText).or(() -> getFieldTemplateText(document.get())).map(Text::getTextContent);
			}
		} catch (RuntimeException e) {
			LOG.warn("Unexpected Error on extracting NachrichtText: {}", e.getMessage(), e);
		}
		return Optional.empty();
	}

	Optional<Document> parseXml(File xmlFile) {
		try {
			return Optional.of(DocumentBuilderFactory.newDefaultInstance().newDocumentBuilder().parse(xmlFile));
		} catch (SAXException | IOException | ParserConfigurationException e) {
			LOG.warn("XML-Parsing error on extracting Nachricht-Text: {}", e.getMessage(), e);
		}
		return Optional.empty();
	}

	Optional<Text> getNachrichtenText(Document document) {
		return evaluateXPath(document, NACHRICHTEN_TEXT_EXPRESSION);
	}

	Optional<Text> getFieldTemplateText(Document document) {
		return evaluateXPath(document, FIELD_TEMPLATE_EXPRESSION);
	}

	Optional<Text> evaluateXPath(Document document, String xpathExpression) {
		try {
			var expr = XPathFactory.newInstance().newXPath().compile(xpathExpression);
			return Optional.ofNullable(expr.evaluate(document, XPathConstants.NODE)).map(Text.class::cast);
		} catch (ClassCastException e) {
			LOG.warn("Error on extraction Nachricht-Text. XPath return unexpected Type.", e);
		} catch (XPathExpressionException e) {
			LOG.warn("Cannot evaluate XPath: {}", xpathExpression, e);
		}
		return Optional.empty();
	}

	Optional<File> getXMLFile(SmartDocumentsResponse response) {
		return getSmartDocumentsFile(response, FILE_TYPE_XML)
				.map(SmartDocumentFile::getDocument)
				.map(SmartDocumentDocument::getData);
	}

	Optional<SmartDocumentFile> getSmartDocumentsFile(SmartDocumentsResponse response, String fileType) {
		return response.getFile().stream()
				.filter(file -> file.getOutputFormat().equals(fileType))
				.findFirst();
	}

	SmartDocumentsRequest createRequest(BescheidRequest request, Vorgang vorgang) {
		return logRequest(SmartDocumentsRequest.builder()
				.smartDocument(buildSDSection(vorgang))
				.customerData(
						CustomerData.builder()
								.vorgang(vorgang)
								.userData(buildUserData(request.getCreateFor()))
								.bescheid(BescheidData.builder().bescheidVom(request.getBescheidVom()).genehmigt(request.isGenehmigt()).build())
								.build())

				.build());
	}

	private UserData buildUserData(UserProfile userProfile) {
		return UserData.builder()
				.firstName(userProfile.getFirstName())
				.lastName(userProfile.getLastName())
				.email(userProfile.getEmail())
				.build();
	}

	private SmartDocumentsRequest.SmartDocument buildSDSection(Vorgang vorgang) {
		return SmartDocumentsRequest.SmartDocument.builder().selection(Selection.builder()
				.templateGroup(properties.getTemplateGroup())
				.template(vorgang.getVorgangName())
				.build())
				.build();
	}

	private SmartDocumentsRequest logRequest(SmartDocumentsRequest request) {
		LOG.debug(() -> {
			var ojMapper = new ObjectMapper();
			try {
				return ojMapper.writeValueAsString(request);
			} catch (JsonProcessingException e) {
				LOG.warn("Json Processing Exception on logging request.", e);
				return request.toString();
			}
		});

		return request;
	}

	@Builder
	@Getter
	@NoArgsConstructor
	@AllArgsConstructor
	static class SmartDocumentsResponse {
		Collection<SmartDocumentFile> file;

		@Builder
		@Getter
		@NoArgsConstructor
		@AllArgsConstructor
		static class SmartDocumentFile {
			private String filename;
			private SmartDocumentDocument document;
			private String outputFormat;
		}

		@Builder
		@Getter
		@NoArgsConstructor
		@AllArgsConstructor
		static class SmartDocumentDocument {
			@JsonDeserialize(using = FileDataDeserializer.class)
			private File data;
		}
	}

}
