/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid.smartdocuments;

import java.util.Objects;

import org.apache.hc.client5.http.auth.AuthScope;
import org.apache.hc.client5.http.auth.CredentialsProvider;
import org.apache.hc.client5.http.impl.auth.CredentialsProviderBuilder;
import org.apache.hc.client5.http.impl.classic.CloseableHttpClient;
import org.apache.hc.client5.http.impl.classic.HttpClientBuilder;
import org.apache.hc.client5.http.impl.classic.HttpClients;
import org.apache.hc.client5.http.impl.io.PoolingHttpClientConnectionManagerBuilder;
import org.apache.hc.client5.http.impl.routing.DefaultProxyRoutePlanner;
import org.apache.hc.client5.http.io.HttpClientConnectionManager;
import org.apache.hc.client5.http.ssl.DefaultClientTlsStrategy;
import org.apache.hc.core5.http.HttpHost;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.boot.ssl.NoSuchSslBundleException;
import org.springframework.boot.ssl.SslBundles;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.HttpHeaders;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.web.client.RestClient;

import lombok.AllArgsConstructor;

@AllArgsConstructor
@Configuration
@ConditionalOnProperty("ozgcloud.bescheid.smart-documents.url")
class SmartDocumentsConfiguration {

	private final SmartDocumentsProperties properties;
	private final SslBundles sslBundles;

	@Bean("smartDocuments")
	RestClient smartDocumentsRestClient() {
		return RestClient.builder()
				.requestFactory(new HttpComponentsClientHttpRequestFactory(buildHttpClient()))
				.baseUrl(properties.getUrl())
				.defaultHeaders(this::addBasicAuthenticationIfConfigured)
				.build();
	}

	CloseableHttpClient buildHttpClient() {
		if (properties.getProxy() != null) {
			return createHttpClientUsingProxy();
		}
		return createNoProxyHttpClient();
	}

	private HttpClientBuilder createDefaultHttpClientBuilder() {
		return HttpClients.custom()
				.setConnectionManager(createConnectionManagerWithClientCertificateIfConfigured());
	}

	CloseableHttpClient createNoProxyHttpClient() {
		return createDefaultHttpClientBuilder().build();
	}

	CloseableHttpClient createHttpClientUsingProxy() {
		return createDefaultHttpClientBuilder()
				.setRoutePlanner(createProxyRoutePlanner())
				.setDefaultCredentialsProvider(createProxyCredentialsProvider())
				.build();
	}

	DefaultProxyRoutePlanner createProxyRoutePlanner() {
		return new DefaultProxyRoutePlanner(new HttpHost(properties.getProxy().getHost(), properties.getProxy().getPort()));
	}

	private CredentialsProvider createProxyCredentialsProvider() {
		var host = properties.getProxy().getHost();
		var port = properties.getProxy().getPort();
		var username = properties.getProxy().getUsername();
		var password = properties.getProxy().getPassword().toCharArray();

		return CredentialsProviderBuilder.create().add(new AuthScope(host, port), username, password).build();
	}

	void addBasicAuthenticationIfConfigured(HttpHeaders headers) {
		if (Objects.nonNull(properties.getBasicAuth())) {
			headers.setBasicAuth(properties.getBasicAuth().getUsername(), properties.getBasicAuth().getPassword());
		}
	}

	HttpClientConnectionManager createConnectionManagerWithClientCertificateIfConfigured() {
		var builder = PoolingHttpClientConnectionManagerBuilder.create();
		try {
			var sslBundleName = properties.getSslBundleName();
			if (Objects.nonNull(sslBundleName)) {
				builder.setTlsSocketStrategy(new DefaultClientTlsStrategy(sslBundles.getBundle(sslBundleName).createSslContext()));
			}
			return builder.build();
		} catch (NoSuchSslBundleException e) {
			return builder.build();
		}
	}

}
