package de.ozgcloud.document.bescheid.smartdocuments;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import javax.net.ssl.SSLContext;

import org.apache.hc.client5.http.auth.AuthScope;
import org.apache.hc.client5.http.auth.UsernamePasswordCredentials;
import org.apache.hc.client5.http.impl.classic.CloseableHttpClient;
import org.apache.hc.client5.http.impl.classic.HttpClients;
import org.apache.hc.client5.http.impl.io.PoolingHttpClientConnectionManager;
import org.apache.hc.client5.http.impl.io.PoolingHttpClientConnectionManagerBuilder;
import org.apache.hc.client5.http.protocol.HttpClientContext;
import org.apache.hc.core5.http.HttpException;
import org.apache.hc.core5.http.HttpHost;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;
import org.springframework.boot.ssl.NoSuchSslBundleException;
import org.springframework.boot.ssl.SslBundle;
import org.springframework.boot.ssl.SslBundles;
import org.springframework.http.HttpHeaders;

import de.ozgcloud.document.bescheid.smartdocuments.SmartDocumentsProperties.ProxyConfiguration;

class SmartDocumentsConfigurationTest {

	@Spy
	@InjectMocks
	private SmartDocumentsConfiguration smartDocumentsConfiguration;

	@Mock
	private SmartDocumentsProperties properties;
	@Mock
	private SslBundles sslBundles;

	@Nested
	class TestBuildHttpClient {
		private final CloseableHttpClient proxyHttpClient = HttpClients.createDefault();
		private final CloseableHttpClient nonProxyHttpClient = HttpClients.createDefault();

		@Test
		void shouldReturnProxyHttpClient() {
			doReturn(proxyHttpClient).when(smartDocumentsConfiguration).createHttpClientUsingProxy();
			when(properties.getProxy()).thenReturn(new ProxyConfiguration());

			var httpClient = smartDocumentsConfiguration.buildHttpClient();

			assertThat(httpClient).isSameAs(proxyHttpClient);
		}

		@Test
		void shouldReturnNonProxyHttpClient() {
			doReturn(nonProxyHttpClient).when(smartDocumentsConfiguration).createNoProxyHttpClient();
			when(properties.getProxy()).thenReturn(null);

			var httpClient = smartDocumentsConfiguration.buildHttpClient();

			assertThat(httpClient).isSameAs(nonProxyHttpClient);
		}
	}

	@Nested
	class TestCreateProxyRoutePlanner {

		private final String host = "test-proxy.local";
		private final int port = 8080;
		private final ProxyConfiguration proxyConfiguration = new ProxyConfiguration();

		@BeforeEach
		void setUp() {
			proxyConfiguration.setHost(host);
			proxyConfiguration.setPort(port);
			when(properties.getProxy()).thenReturn(proxyConfiguration);
		}

		@Test
		void shouldCreateWithProxySettings() throws HttpException {
			var routePlanner = smartDocumentsConfiguration.createProxyRoutePlanner();

			var proxyHost = routePlanner.determineRoute(new HttpHost("http://any"), new HttpClientContext()).getProxyHost();

			assertThat(proxyHost).extracting(
					HttpHost::getHostName,
					HttpHost::getPort
			).containsExactly(host, port);
		}
	}

	@Nested
	class TestCreateProxyCredentialsProvider {

		private final String user = "max";
		private final String password = "max2";
		private final String host = "test-proxy.local";
		private final int port = 8080;
		private final ProxyConfiguration proxyConfiguration = new ProxyConfiguration();

		@BeforeEach
		void setUp() {
			proxyConfiguration.setHost(host);
			proxyConfiguration.setPort(port);
			proxyConfiguration.setUsername(user);
			proxyConfiguration.setPassword(password);
			when(properties.getProxy()).thenReturn(proxyConfiguration);
		}

		@Test
		void shouldCreateWithUserAndPassword() {
			var credentialsProvider = smartDocumentsConfiguration.createProxyCredentialsProvider();

			var credentialsForProxy = (UsernamePasswordCredentials) credentialsProvider.getCredentials(new AuthScope(host, port), null);

			assertThat(credentialsForProxy).extracting(
							UsernamePasswordCredentials::getUserName,
							UsernamePasswordCredentials::getUserPassword)
					.containsExactly(user, password.toCharArray());
		}
	}

	@Nested
	class TestAddBasicAuthenticationIfConfigured {

		@Test
		void shouldNotAddBasicAuthentication() {
			when(properties.getBasicAuth()).thenReturn(null);
			var headers = new HttpHeaders();

			smartDocumentsConfiguration.addBasicAuthenticationIfConfigured(headers);

			assertThat(headers.get(HttpHeaders.AUTHORIZATION)).isNull();
		}

		@Test
		void shouldAddBasicAuthentication() {
			when(properties.getBasicAuth()).thenReturn(UsernamePasswordTestFactory.create());
			var headers = new HttpHeaders();

			smartDocumentsConfiguration.addBasicAuthenticationIfConfigured(headers);

			assertThat(headers.get(HttpHeaders.AUTHORIZATION))
					.isNotNull()
					.containsExactly(UsernamePasswordTestFactory.getAsBasicAuthenticationHeader());
		}
	}

	@Nested
	class TestCreateConnectionManagerWithClientCertificateIfConfigured {

		private MockedStatic<PoolingHttpClientConnectionManagerBuilder> staticBuilder;
		@Mock
		private PoolingHttpClientConnectionManagerBuilder builder;
		@Mock
		private SslBundle sslBundle;
		@Mock
		private SSLContext sslContext;

		private final String bundleName = "smartdocuments_bundle";
		private final PoolingHttpClientConnectionManager connectionManager = new PoolingHttpClientConnectionManager();

		@BeforeEach
		void setUp() {
			staticBuilder = mockStatic(PoolingHttpClientConnectionManagerBuilder.class);
			staticBuilder.when(PoolingHttpClientConnectionManagerBuilder::create).thenReturn(builder);
			when(builder.build()).thenReturn(connectionManager);
		}

		@AfterEach
		void tearDown() {
			staticBuilder.close();
		}

		@Test
		void shouldCreateWithBuilder() {
			smartDocumentsConfiguration.createConnectionManagerWithClientCertificateIfConfigured();

			staticBuilder.verify(() -> PoolingHttpClientConnectionManagerBuilder.create());
		}

		@Nested
		class OnMissingClientCertificateConfiguration {

			@Test
			void shouldCreateWithoutTlsStrategyOnMissingBundleName() {
				when(properties.getSslBundleName()).thenReturn(null);

				var created = smartDocumentsConfiguration.createConnectionManagerWithClientCertificateIfConfigured();

				verify(builder, never()).setTlsSocketStrategy(any());
				assertThat(created).isSameAs(connectionManager);

			}

			@Test
			void shouldCreateWithoutTlsStrategyOnMissingBundleConfiguration() {
				when(properties.getSslBundleName()).thenReturn(bundleName);
				when(sslBundles.getBundle(anyString())).thenThrow(new NoSuchSslBundleException("some_bundle", "error message"));

				var created = smartDocumentsConfiguration.createConnectionManagerWithClientCertificateIfConfigured();

				verify(builder, never()).setTlsSocketStrategy(any());
				assertThat(created).isSameAs(connectionManager);
			}

		}

		@Nested
		class OnExistingClientCertificateConfiguration {

			@BeforeEach
			void setUp() {
				when(properties.getSslBundleName()).thenReturn(bundleName);
				when(sslBundles.getBundle(any())).thenReturn(sslBundle);
				when(sslBundle.createSslContext()).thenReturn(sslContext);
			}

			@Test
			void shouldGetSslBundle() {
				smartDocumentsConfiguration.createConnectionManagerWithClientCertificateIfConfigured();

				verify(sslBundles).getBundle(bundleName);
			}

			@Test
			void shouldCreateWithTlsStrategy() {
				var created = smartDocumentsConfiguration.createConnectionManagerWithClientCertificateIfConfigured();

				verify(builder).setTlsSocketStrategy(any());
				assertThat(created).isSameAs(connectionManager);
			}
		}

	}
}